import json
import pathlib
import typing

import click

from unitelabs.cdk import utils
from unitelabs.cdk.config import get_connector_config
from unitelabs.cdk.main import load


@click.group(context_settings=dict(show_default=True))
def config() -> click.Group:
    """Configure a connector."""


@config.command()
@click.option(
    "--app",
    type=str,
    metavar="IMPORT",
    default=utils.find_factory,
    show_default=False,
    help="The application factory function to load, in the form 'module:name'.",
)
@click.option(
    "-p",
    "--path",
    type=click.Path(
        exists=False,
        dir_okay=False,
        writable=True,
        resolve_path=True,
        path_type=pathlib.Path,
    ),
    default=pathlib.Path("./schema.json"),
    help="Path to the configuration schema file.",
)
@utils.coroutine
async def schema(app, path: pathlib.Path) -> None:  # noqa: ANN001
    """Create a configuration jsonschema."""
    await load(app)
    config = get_connector_config()
    with path.open("w") as file:
        json.dump(config.schema(), file, indent=2)


@config.command()
@click.option(
    "--app",
    type=str,
    metavar="IMPORT",
    default=utils.find_factory,
    show_default=False,
    help="The application factory function to load, in the form 'module:name'.",
)
@click.option(
    "-f",
    "--field",
    type=str,
    required=False,
    help="The name of the field in the schema to get more information about, otherwise the entire schema is shown.",
)
@utils.coroutine
async def show(app, field: typing.Optional[str] = None) -> None:  # noqa: ANN001
    """Visualize the configuration options."""
    await load(app)
    config = get_connector_config()
    description = config.describe(field)

    from rich.console import Console
    from rich.table import Column, Table

    table = Table(
        Column("Field", justify="left"),
        Column("Type", justify="right"),
        Column("Description", justify="right"),
        Column("Example", justify="right"),
        title=f"{config.__name__} Definition",
        show_lines=True,
    )

    for name, values in description.items():
        if not isinstance(values, dict):
            table.add_row(field, *description.values())
            break
        if "values" not in values:
            table.add_row(name, *values.values())
        else:
            table.add_row(name, values["type"], values["description"], values["default"])

    console = Console()
    console.print(table)


@config.command()
@click.option(
    "--app",
    type=str,
    metavar="IMPORT",
    default=utils.find_factory,
    show_default=False,
    help="The application factory function to load, in the form 'module:name'.",
)
@click.option(
    "--path",
    type=click.Path(
        exists=False,
        dir_okay=False,
        writable=True,
        resolve_path=True,
        path_type=pathlib.Path,
    ),
    default=pathlib.Path("./config.json"),
    help="Path to the configuration file.",
)
@utils.coroutine
async def create(app, path: pathlib.Path) -> None:  # noqa: ANN001
    """Create a configuration file."""
    await load(app)
    config = get_connector_config()
    config().dump(path)


if __name__ == "__main__":
    config()



import inspect

import typing_extensions as typing

import sila

from .. import utils
from ..commands import IntermediateResponses, Parameters, Responses
from ..common import Decorator
from ..common.errors import define_error
from ..metadata import Metadatum

if typing.TYPE_CHECKING:
    from ..common import Feature


class ObservableCommand(Decorator):
    """
    Any command for which observing the progress of execution is possible or does make sense.

    Args:
      name: Human readable name for the command. By default, this is
        automatically inferred by the name of the decorated method.
      identifier: Unique identifier of the command. By default, this
        equals the `name` without spaces.
      errors: A list of defined errors that may occur during command
        execution.

    Examples:
      Convert a feature method into an observable command:
      >>> class MyFeature(sila.Feature):
      ...   @sila.ObservableCommand
      ...   @sila.Response("Response A")
      ...   @sila.Response("Response B")
      ...   async def my_command(self, param_a: str, param_b: int) -> tuple[str, int]:
      ...     \"\"\"
      ...     Describe what your command does.
      ...     .. parameter:: Describe the purpose of param_a.
      ...     .. parameter:: Describe the purpose of param_b.
      ...     .. return:: The input parameter directly returned.
      ...     .. return:: Describe the purpose of param_b.
      ...     \"\"\"
      ...     return param_a, param_b
    """

    @typing.override
    def attach(self, feature: "Feature") -> bool:
        if not super().attach(feature):
            return False

        docs = utils.parse_docs(inspect.getdoc(self._function))

        self._name = self._name or utils.humanize(self._function.__name__)
        self._identifier = self._identifier or self._name.replace(" ", "")
        self._description = docs.get("default", "")

        self._parameters = Parameters.from_signature(feature, self._function).elements
        self._responses = Responses.from_signature(feature, self._function).elements
        self._intermediate_responses = IntermediateResponses.from_signature(feature, self._function).elements

        self._handler = sila.server.ObservableCommand(
            identifier=self._identifier,
            display_name=self._name,
            description=self._description,
            function=self.execute,
            errors={Error.identifier: Error for error in self._errors if (Error := define_error(error))},
            parameters=self._parameters,
            responses=self._responses,
            intermediate_responses=self._intermediate_responses,
            feature=feature,
        )
        self._metadata = Metadatum._infer_metadata(self)

        return True



import dataclasses
import typing

import pytest

from unitelabs.cdk.config import Config
from unitelabs.cdk.config.schema import InvalidSchemaFieldError, Schema, describe, get_type_str


@dataclasses.dataclass
class NestedRef(Config):
    test_ref_value: str = "REF_VALUE"
    """A nested property."""


@dataclasses.dataclass
class DescribableConfig(Config):
    test_ref: NestedRef = dataclasses.field(default_factory=NestedRef)
    """A nested config object."""
    test_property: str = "PROPERTY_VALUE"
    """A string property."""


TEST_SCHEMA = {
    "properties": {
        "test_ref": {"$ref": "#/$defs/NestedRef"},
        "test_property": {
            "type": "string",
            "default": "PROPERTY_VALUE",
            "title": "Test Property",
            "description": "A string property.",
        },
    },
    "$defs": {
        "NestedRef": {
            "properties": {
                "test_ref_value": {
                    "type": "string",
                    "default": "REF_VALUE",
                    "title": "Test Ref Value",
                    "description": "A nested property.",
                }
            },
            "type": "object",
            "title": "NestedRef",
        },
    },
    "type": "object",
    "title": "DescribableConfig",
    "$schema": "https://json-schema.org/draft/2020-12/schema",
}


class TestGet:
    def test_should_get_valid_reference(self):
        schema = Schema(TEST_SCHEMA)
        test_ref = schema.get("test_ref")
        assert isinstance(test_ref, Schema)
        assert test_ref.definition == {
            "properties": {
                "test_ref_value": {
                    "type": "string",
                    "default": "REF_VALUE",
                    "title": "Test Ref Value",
                    "description": "A nested property.",
                }
            },
            "type": "object",
            "title": "NestedRef",
        }
        assert test_ref._schema == TEST_SCHEMA["$defs"]["NestedRef"]

    def test_should_get_valid_property(self):
        schema = Schema(TEST_SCHEMA)
        test_value = schema.get("test_property")
        assert isinstance(test_value, Schema)
        assert test_value.definition == {
            "type": "string",
            "default": "PROPERTY_VALUE",
            "title": "Test Property",
            "description": "A string property.",
        }
        assert test_value._schema == TEST_SCHEMA["properties"]["test_property"]

    def test_should_raise_invalid_schema_field_error(self):
        schema = Schema(TEST_SCHEMA)
        with pytest.raises(InvalidSchemaFieldError):
            schema.get("invalid_field")


class TestDescribe:
    def test_should_describe_config(self):
        schema = DescribableConfig.schema()
        data = describe(DescribableConfig, Schema(schema))
        assert data == {
            "test_ref": {
                "default": "",
                "type": "NestedRef",
                "description": "For more information use `config show --field test_ref`.",
                "values": {
                    "test_ref_value": {
                        "type": "str",
                        "description": "A nested property.",
                        "default": "'REF_VALUE'",
                    }
                },
            },
            "test_property": {"type": "str", "description": "A string property.", "default": "'PROPERTY_VALUE'"},
        }


class TestGetTypeStr:
    @pytest.mark.parametrize(
        (
            "type_",
            "expected",
        ),
        [
            pytest.param(int, "int", id="int"),
            pytest.param(str, "str", id="str"),
            pytest.param(list[int], "list[int]", id="list[int]"),
            pytest.param(list[str], "list[str]", id="list[str]"),
            pytest.param(dict[str, int], "dict[str, int]", id="dict[str, int]"),
            pytest.param(dict[str, str], "dict[str, str]", id="dict[str, str]"),
            pytest.param(set[int], "set[int]", id="set[int]"),
            pytest.param(set[str], "set[str]", id="set[str]"),
            pytest.param(list[DescribableConfig], "list[DescribableConfig]", id="list[DescribableConfig]"),
            pytest.param(set[DescribableConfig], "set[DescribableConfig]", id="set[DescribableConfig]"),
            pytest.param(
                list[typing.Union[DescribableConfig, int]],
                "list[DescribableConfig | int]",
                id="list[DescribableConfig | int]",
            ),
            pytest.param(typing.Annotated[int, "test"], "int", id="int"),
            pytest.param(typing.Literal[1, 2, 3], "1 | 2 | 3", id="Literal[1, 2, 3]"),
            pytest.param(typing.Literal["a", "b", "c"], "'a' | 'b' | 'c'", id="Literal['a', 'b', 'c']"),
        ],
    )
    def test_should_get_type_str(self, type_, expected):
        assert get_type_str(type_) == expected



import collections.abc
import inspect

import typing_extensions as typing

import sila
from sila.server import Native

from .. import utils
from ..common import Decorator
from ..common.errors import define_error
from ..data_types import infer
from ..metadata import Metadatum

if typing.TYPE_CHECKING:
    from ..common import Feature


class UnobservableProperty(Decorator):
    """A property describes certain aspects of a SiLA server that do not require an action on the SiLA server."""

    @typing.override
    def attach(self, feature: "Feature") -> bool:
        if not super().attach(feature):
            return False

        self._name = self._name or utils.humanize(self._function.__name__.removeprefix("get_"))
        self._identifier = self._identifier or self._name.replace(" ", "")
        self._description = inspect.getdoc(self._function) or ""

        type_hint = inspect.signature(self._function).return_annotation
        if type_hint is inspect._empty:
            type_hint = Native

        data_type = infer(type_hint, feature)

        self._responses = {
            self._identifier: sila.Element(identifier=self._identifier, display_name=self._name, data_type=data_type)
        }

        self._handler = sila.server.UnobservableProperty(
            identifier=self._identifier,
            display_name=self._name,
            description=self._description,
            function=self.execute,
            errors={Error.identifier: Error for error in self._errors if (Error := define_error(error))},
            data_type=data_type,
            feature=feature,
        )
        self._metadata = Metadatum._infer_metadata(self)

        return True

    @typing.override
    async def _execute(self, function: collections.abc.Callable) -> Native:
        responses: Native = function()

        if inspect.isawaitable(responses):
            responses = await responses

        return {self._identifier: responses}



import dataclasses
import pathlib
import uuid

import pydantic
import pydantic_core
import typing_extensions as typing
from pydantic import Field
from pydantic.json_schema import JsonSchemaValue

import sila

from ..sila.utils import parse_version
from .config import Config, ConfigurationError

URIString = typing.Annotated[
    str,
    pydantic.WithJsonSchema({"type": "string", "format": "uri"}),
]

UUIDString = typing.Annotated[
    str,
    pydantic.WithJsonSchema({"type": "string", "format": "uuid"}),
]


def read_bytes_if_path_and_exists(path: typing.Union[str, pathlib.Path, bytes, None]) -> typing.Optional[bytes]:
    """
    Read the byte-contents of the given `path`, if it is a path or a string-representation of a path.

    If the path-string's resolved `Path` does not exist, it is treated as a base64-encoded ASCII-string
    and decoded to bytes.

    Args:
      path: A string-representation of a path, or a path, from which to read the contents,
        or base64-encoded ASCII-string, which is decoded to bytes,
        or a bytestring or None, which are returned as-is.

    Returns:
      The byte-contents from `path` or None.

    Raises:
      FileNotFoundError: If `path` is a valid path but does not exist.
    """

    if not isinstance(path, (str, pathlib.Path)):
        return path

    if isinstance(path, str):
        path_str = path
        path = pathlib.Path(path).resolve()

    try:
        if not path.resolve().exists():
            msg = f"File at path '{path.resolve()}' not found."
            raise FileNotFoundError(msg)
    except OSError as err:
        if not err.errno == 63:  # File name too long
            raise
        # assumed to be a base64-encoded string of the certificate/key
        return path_str.encode("ascii")

    return path.read_bytes()


@dataclasses.dataclass
class SiLAServerConfig(sila.server.ServerConfig, Config):
    """Configuration for a SiLA server."""

    certificate_chain: typing.Union[str, pathlib.Path, bytes, None] = None
    """
    A path to, or the bytestring contents of, the PEM-encoded certificate chain, or `None`
    if no certificate chain should be used.
    Note: TLS must be set to True to activate encryption with this certificate.
    """

    private_key: typing.Union[str, pathlib.Path, bytes, None] = None
    """
    A path to, or the bytestring contents of, the PEM-encoded private key, or `None` if no
    private key should be used.
    Note: TLS must be set to True to activate encryption with this key.
    """

    options: dict = dataclasses.field(default_factory=dict)
    uuid: UUIDString = dataclasses.field(default_factory=lambda: str(uuid.uuid4()))
    name: typing.Annotated[str, Field(max_length=255)] = "SiLA Server"
    vendor_url: URIString = "https://sila-standard.com"

    @classmethod
    def __get_pydantic_json_schema__(
        cls, core_schema: pydantic_core.core_schema.CoreSchema, handler: pydantic.GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        json_schema = handler(core_schema)
        json_schema["required"] = ["uuid"]
        return json_schema

    def __post_init__(self):
        self.certificate_chain = read_bytes_if_path_and_exists(self.certificate_chain)
        self.private_key = read_bytes_if_path_and_exists(self.private_key)
        self.version = parse_version(self.version)


@dataclasses.dataclass
class CloudServerConfig(sila.server.CloudServerConfig, Config):
    """Configuration for a gRPC Cloud Server."""

    port: typing.Annotated[int, Field(ge=1, le=65_535)] = 50_000

    certificate_chain: typing.Union[str, pathlib.Path, bytes, None] = None
    """
    A path to, or the bytestring contents of, the PEM-encoded certificate chain, or `None` if no
    certificate chain should be used.
    Note: TLS must be set to True to activate encryption with this certificate.
    """

    private_key: typing.Union[str, pathlib.Path, bytes, None] = None
    """
    A path to, or the bytestring contents of, the PEM-encoded private key, or `None` if no
    private key should be used.
    Note: TLS must be set to True to activate encryption with this key.
    """
    options: dict = dataclasses.field(default_factory=dict)

    def __post_init__(self):
        self.certificate_chain = read_bytes_if_path_and_exists(self.certificate_chain)
        self.private_key = read_bytes_if_path_and_exists(self.private_key)


@dataclasses.dataclass
class ConnectorBaseConfig(Config):
    """Base configuration for a UniteLabs SiLA2 Connector."""

    sila_server: SiLAServerConfig = dataclasses.field(default_factory=SiLAServerConfig)
    cloud_server_endpoint: typing.Optional[CloudServerConfig] = dataclasses.field(default_factory=CloudServerConfig)
    logging: typing.Optional[dict] = dataclasses.field(default=None)

    def __post_init__(self):
        if isinstance(self.sila_server, dict):
            self.sila_server = SiLAServerConfig(**self.sila_server)
        if isinstance(self.cloud_server_endpoint, dict):
            self.cloud_server_endpoint = CloudServerConfig(**self.cloud_server_endpoint)


def get_connector_config() -> type[ConnectorBaseConfig]:
    """Get the current connector configuration."""
    derived_configs = {c for c in ConnectorBaseConfig.__subclasses__() if c.__name__ != "ConnectorBaseConfig"}
    if len(derived_configs) > 1 and len({c.__name__ for c in derived_configs}) > 1:
        msg = (
            f"Multiple configurations ({', '.join([c.__name__ for c in derived_configs])}) found. "
            "Please ensure only one subclass of ConnectorBaseConfig exists."
        )
        raise ConfigurationError(msg)

    return derived_configs.pop() if derived_configs else ConnectorBaseConfig


__all__ = [
    "CloudServerConfig",
    "ConnectorBaseConfig",
    "SiLAServerConfig",
    "get_connector_config",
]



import unittest.mock

import pytest

from sila import Element, Integer, String
from unitelabs.cdk.sila.commands.responses import Response, Responses


class TestFromSignature:
    # Verify that the method raises when typing annotations are unkown.
    async def test_raises_when_annotation_unknown(self):
        # Initialize the method
        feature = unittest.mock.Mock()

        def handler() -> complex: ...

        # Infer responses
        with pytest.raises(TypeError, match=r"Unable to identify SiLA type from annotation 'complex' in .+\.handler\."):
            Responses.from_signature(feature, handler)

    # Verify that the method raises when typing annotations are invalid.
    async def test_raises_when_annotation_invalid(self):
        # Initialize the method
        feature = unittest.mock.Mock()

        def handler() -> list: ...

        # Infer responses
        with pytest.raises(TypeError, match=r"Unable to identify SiLA type from annotation 'list' in .+\.handler\."):
            Responses.from_signature(feature, handler)

    # Infers responses with typing annotations missing.
    def test_infer_responses_annotation_missing(self):
        # Initialize the method
        feature = unittest.mock.Mock()

        def handler(): ...

        # Infer responses
        responses = Responses.from_signature(feature, handler)

        # Assert that the method returns the correct value
        assert responses.elements == {}

    # Infers responses with return none.
    def test_infer_responses_return_none(self):
        # Initialize the method
        feature = unittest.mock.Mock()

        def handler() -> None: ...

        # Infer responses
        responses = Responses.from_signature(feature, handler)

        # Assert that the method returns the correct value
        assert responses.elements == {}

    # Infers responses with signature.
    def test_infer_responses_signature(self):
        # Initialize the method
        feature = unittest.mock.Mock()

        def handler() -> str: ...

        # Infer responses
        with pytest.warns(match=r"No name found for Responses 0 in .+\.handler, defaulting to 'Unnamed 0'\."):
            responses = Responses.from_signature(feature, handler)

        # Assert that the method returns the correct value
        assert responses.elements == {
            "Unnamed0": Element(identifier="Unnamed0", display_name="Unnamed 0", description="", data_type=String),
        }

    # Infers responses with tuple.
    def test_infer_responses_tuple(self):
        # Initialize the method
        feature = unittest.mock.Mock()

        def handler() -> tuple[str, int]: ...

        # Infer responses
        with (
            pytest.warns(match=r"No name found for Responses 0 in .+\.handler, defaulting to 'Unnamed 0'\."),
            pytest.warns(match=r"No name found for Responses 1 in .+\.handler, defaulting to 'Unnamed 1'\."),
        ):
            responses = Responses.from_signature(feature, handler)

        # Assert that the method returns the correct value
        assert responses.elements == {
            "Unnamed0": Element(identifier="Unnamed0", display_name="Unnamed 0", description="", data_type=String),
            "Unnamed1": Element(identifier="Unnamed1", display_name="Unnamed 1", description="", data_type=Integer),
        }

    # Infers responses with decorators.
    def test_infer_responses_decorator(self):
        # Initialize the method
        feature = unittest.mock.Mock()

        @Response("Variable A")
        @Response("Variable B")
        def handler() -> tuple[str, int]:
            """
            .. return:: Explanation of variable a.
            .. return:: Explanation of variable b.
            """

            raise NotImplementedError

        # Infer responses
        responses = Responses.from_signature(feature, handler)

        # Assert that the method returns the correct value
        assert responses.elements == {
            "VariableA": Element(
                identifier="VariableA",
                display_name="Variable A",
                description="Explanation of variable a.",
                data_type=String,
            ),
            "VariableB": Element(
                identifier="VariableB",
                display_name="Variable B",
                description="Explanation of variable b.",
                data_type=Integer,
            ),
        }

    # Infers responses with documentation.
    def test_infer_responses_documentation(self):
        # Initialize the method
        feature = unittest.mock.Mock()

        @Response("Variable A")
        @Response("Variable B")
        def handler() -> tuple[str, int]:
            """
            .. return:: Explanation of variable a.
            .. return:: Explanation of variable b.
                :name: Override B
            """

            raise NotImplementedError

        # Infer responses
        responses = Responses.from_signature(feature, handler)

        # Assert that the method returns the correct value
        assert responses.elements == {
            "VariableA": Element(
                identifier="VariableA",
                display_name="Variable A",
                description="Explanation of variable a.",
                data_type=String,
            ),
            "OverrideB": Element(
                identifier="OverrideB",
                display_name="Override B",
                description="Explanation of variable b.",
                data_type=Integer,
            ),
        }

    async def test_that_decorator_overrides_docs_description(self):
        # Initialize the method
        feature = unittest.mock.Mock()

        @Response("Variable A", "Overriding explanation of variable a.")
        @Response("Variable B")
        def handler() -> tuple[str, int]:
            """
            .. return:: Explanation of variable a.
            .. return:: Explanation of variable b.
            """

            raise NotImplementedError

        # Infer responses
        responses = Responses.from_signature(feature, handler)

        # Assert that the method returns the correct value
        assert responses.elements == {
            "VariableA": Element(
                identifier="VariableA",
                display_name="Variable A",
                description="Overriding explanation of variable a.",
                data_type=String,
            ),
            "VariableB": Element(
                identifier="VariableB",
                display_name="Variable B",
                description="Explanation of variable b.",
                data_type=Integer,
            ),
        }

    # Verify that the method warns when more documentations than arguments are provided.
    async def test_warns_when_more_documentations_than_arguments(self):
        # Initialize the method
        feature = unittest.mock.Mock()

        def handler() -> str:
            """
            .. return:: Explanation of variable a.
                :name: Override A
            .. return:: Explanation of variable b.
                :name: Override B
            """

            raise NotImplementedError

        # Infer parameters
        with pytest.warns(match=r"More documented items than annotations in .+\.handler, using only the first 1\."):
            Responses.from_signature(feature, handler)

    # Verify that the method warns when more decorators than arguments are provided.
    async def test_warns_when_more_decorators_than_arguments(self):
        # Initialize the method
        feature = unittest.mock.Mock()

        @Response("Variable A")
        @Response("Variable B")
        def handler() -> str: ...

        # Infer parameters
        with pytest.warns(match=r"More decorators than annotations in .+\.handler, using only the first 1\."):
            Responses.from_signature(feature, handler)



import unittest.mock

import pytest

from sila import Element, Integer, String
from unitelabs.cdk.sila.commands.parameters import Parameter, Parameters


class TestFromSignature:
    # Infers parameters with no parameter.
    def test_infer_parameters_no_parameter(self):
        # Initialize the method
        feature = unittest.mock.Mock()

        def handler(): ...

        # Infer parameters
        parameters = Parameters.from_signature(feature, handler)

        # Assert that the method returns the correct value
        assert parameters.elements == {}

    # Infers parameters with self parameter.
    def test_infer_parameters_self_parameter(self):
        # Initialize the method
        feature = unittest.mock.Mock()

        def handler(self): ...

        # Infer parameters
        parameters = Parameters.from_signature(feature, handler)

        # Assert that the method returns the correct value
        assert parameters.elements == {}

    # Verify that the method raises when typing annotations are missing.
    async def test_raises_when_annotation_missing(self):
        # Initialize the method
        feature = unittest.mock.Mock()

        def handler(self, variable_a, variable_b): ...

        # Infer parameters
        with pytest.raises(TypeError, match=r"Missing type annotation for parameter 'variable_a' in .+\.handler\."):
            Parameters.from_signature(feature, handler)

    # Verify that the method raises when typing annotations are unkown.
    async def test_raises_when_annotation_unknown(self):
        # Initialize the method
        feature = unittest.mock.Mock()

        def handler(self, variable_a: int, variable_b: complex): ...

        # Infer parameters
        with pytest.raises(TypeError, match=r"Unable to identify SiLA type from annotation 'complex' in .+\.handler\."):
            Parameters.from_signature(feature, handler)

    # Verify that the method raises when typing annotations are invalid.
    async def test_raises_when_annotation_invalid(self):
        # Initialize the method
        feature = unittest.mock.Mock()

        def handler(self, variable_a: list, variable_b: str): ...

        # Infer parameters
        with pytest.raises(TypeError, match=r"Unable to identify SiLA type from annotation 'list' in .+\.handler\."):
            Parameters.from_signature(feature, handler)

    # Infers parameters with signature.
    def test_infer_parameters_signature(self):
        # Initialize the method
        feature = unittest.mock.Mock()

        def handler(self, variable_a: str, variable_b: int): ...

        # Infer parameters
        parameters = Parameters.from_signature(feature, handler)

        # Assert that the method returns the correct value
        assert parameters.elements == {
            "variable_a": Element(identifier="VariableA", display_name="Variable A", description="", data_type=String),
            "variable_b": Element(identifier="VariableB", display_name="Variable B", description="", data_type=Integer),
        }

    # Infers parameters with decorators.
    def test_infer_parameters_decorator(self):
        # Initialize the method
        feature = unittest.mock.Mock()

        @Parameter("Override A")
        @Parameter("Override B")
        def handler(self, variable_a: str, variable_b: int): ...

        # Infer parameters
        parameters = Parameters.from_signature(feature, handler)

        # Assert that the method returns the correct value
        assert parameters.elements == {
            "variable_a": Element(
                identifier="OverrideA",
                display_name="Override A",
                description="",
                data_type=String,
            ),
            "variable_b": Element(
                identifier="OverrideB",
                display_name="Override B",
                description="",
                data_type=Integer,
            ),
        }

    # Infers parameters with documentation.
    def test_infer_parameters_documentation(self):
        # Initialize the method
        feature = unittest.mock.Mock()

        @Parameter("Variable A")
        @Parameter("Variable B")
        def handler(self, variable_a: str, variable_b: int):
            """
            .. parameter:: Explanation of variable a.
            .. parameter:: Explanation of variable b.
                :name: Override B
            """

        # Infer parameters
        parameters = Parameters.from_signature(feature, handler)

        # Assert that the method returns the correct value
        assert parameters.elements == {
            "variable_a": Element(
                identifier="VariableA",
                display_name="Variable A",
                description="Explanation of variable a.",
                data_type=String,
            ),
            "variable_b": Element(
                identifier="OverrideB",
                display_name="Override B",
                description="Explanation of variable b.",
                data_type=Integer,
            ),
        }

    async def test_that_decorator_overrides_docs_description(self):
        # Initialize the method
        feature = unittest.mock.Mock()

        @Parameter("Variable A", "Overriding explanation of variable a.")
        @Parameter("Variable B")
        def handler(self, variable_a: str, variable_b: int):
            """
            .. parameter:: Explanation of variable a.
            .. parameter:: Explanation of variable b.
            """

        # Infer parameters
        parameters = Parameters.from_signature(feature, handler)

        # Assert that the method returns the correct value
        assert parameters.elements == {
            "variable_a": Element(
                identifier="VariableA",
                display_name="Variable A",
                description="Overriding explanation of variable a.",
                data_type=String,
            ),
            "variable_b": Element(
                identifier="VariableB",
                display_name="Variable B",
                description="Explanation of variable b.",
                data_type=Integer,
            ),
        }

    # Verify that the method warns when more documentations than arguments are provided.
    async def test_warns_when_more_documentations_than_arguments(self):
        # Initialize the method
        feature = unittest.mock.Mock()

        def handler(self, variable_a: str):
            """
            .. parameter:: Explanation of variable a.
            .. parameter:: Explanation of variable b.
                :name: Override B
            """

        # Infer parameters
        with pytest.warns(match=r"More documented items than annotations in .+\.handler, using only the first 1\."):
            Parameters.from_signature(feature, handler)



from .lock_controller import (
    InvalidLockIdentifier,
    Lock,
    LockController,
    LockIdentifier,
    ServerAlreadyLocked,
    ServerNotLocked,
)

__all__ = [
    "InvalidLockIdentifier",
    "Lock",
    "LockController",
    "LockController",
    "LockIdentifier",
    "ServerAlreadyLocked",
    "ServerNotLocked",
]



import collections.abc
import dataclasses
import inspect

import typing_extensions as typing

from sila.server import Element, Feature

from .. import utils
from .decorator import Decorator, parse_structure

PARAMETERS_ATTRIBUTE = "__parameters"
DOCUMENTATION_KEYWORD = "parameter"


@dataclasses.dataclass
class Parameter(Decorator):
    """
    A decorator class for defining parameters of an SiLA endpoint.

    Attributes:
      name: The name of the decorated parameter.
      description: A description providing more details about the decorated parameter.

    Examples:
      Option 1: Using decorators to define parameters for an endpoint method:
      >>> @Parameter(name="Parameter A", description="A string parameter.")
      >>> @Parameter(name="Parameter B", description="An integer parameter.")
      >>> def example_handler(self, /, param_a: str, param_b: int):
      >>>     ...

      Option 2: Using a docstring to define parameters:
      >>> def example_handler(self, /, param_a: str, param_b: int):
      >>>     \"\"\"
      >>>     .. parameter:: A string parameter.
      >>>       :name: Parameter A
      >>>     .. parameter:: An integer parameter.
      >>>       :name: Parameter B
      >>>     \"\"\"
      >>>     ...
    """

    def __call__(self, function: collections.abc.Callable) -> collections.abc.Callable:  # noqa: D102
        responses = getattr(function, PARAMETERS_ATTRIBUTE, [])
        setattr(function, PARAMETERS_ATTRIBUTE, [self, *responses])

        return function


@dataclasses.dataclass
class Parameters:
    """Represents a SiLA structure containing parameters, inferred from a callable's signature."""

    elements: dict[str, Element] = dataclasses.field(default_factory=dict)

    @classmethod
    def from_signature(
        cls,
        feature: Feature,
        function: collections.abc.Callable,
    ) -> typing.Self:
        """
        Infer and construct parameter elements based on function signature annotations, decorators, and documentation.

        Args:
          feature: The SiLA feature context.
          function: The function to analyze for parameter elements.

        Returns:
          A Parameters object containing the inferred parameter elements.

        Raises:
          TypeError: If any function parameter is missing an annotation or the given annotation is not SiLA compliant.
        """

        signature = inspect.signature(function)
        docs = utils.parse_docs(inspect.getdoc(function)).get(DOCUMENTATION_KEYWORD, [])
        decorators: list[Parameter] = getattr(function, PARAMETERS_ATTRIBUTE, [])

        annotations: typing.Iterable[type] = []
        parameters = list(cls.__get_parameters(signature))
        for i, parameter in enumerate(cls.__get_parameters(signature)):
            if parameter.annotation is inspect.Parameter.empty:
                msg = f"Missing type annotation for parameter '{parameter.name}' in {function.__qualname__}."
                raise TypeError(msg)

            if i < len(decorators):
                decorators[i].name = decorators[i].name or utils.humanize(parameter.name)
            else:
                decorators.append(Parameter(utils.humanize(parameter.name)))

            annotations.append(parameter.annotation)

        structure = parse_structure(
            Parameters,
            feature,
            function,
            annotations,
            decorators,
            docs,
        )

        return cls({parameter.name: element for element, parameter in zip(structure.values(), parameters)})

    @classmethod
    def __get_parameters(cls, signature: inspect.Signature) -> typing.Iterable[inspect.Parameter]:
        """
        Get the relevant parameters from a given signature.

        Args:
          signature: The signature to extract parameters from.

        Returns:
          Parameters that are positional or keyword excluding 'self' and 'cls'.
        """

        return (
            parameter
            for parameter in signature.parameters.values()
            if parameter.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD and parameter.name not in ("self", "cls")
        )

    @classmethod
    def get_mapping(cls, elements: dict[str, Element], function: collections.abc.Callable) -> dict[str, str]:
        """
        Map structure elements to function parameters based on their names.

        Args:
          elements: The structure elements to map.
          function: The function whose parameters are being mapped.

        Returns:
          A mapping of structure element identifiers to function parameter names.
        """

        signature = inspect.signature(function)
        return {
            element.identifier: parameter.name
            for element, parameter in zip(elements.values(), cls.__get_parameters(signature))
        }



import dataclasses
import datetime
import inspect

import typing_extensions as typing

from sila.server import (
    Any,
    Binary,
    Boolean,
    Constrained,
    DataType,
    Date,
    Element,
    Integer,
    List,
    Native,
    Real,
    String,
    Structure,
    Time,
    Timestamp,
    Void,
)

from .. import utils
from ..common import Feature
from .custom_data_type import CustomDataType


def infer(annotation: type, feature: Feature) -> type[DataType]:
    """
    Infer the SiLA data type from a given python type annotation.

    Args:
      annotation: The python type annotation.
      feature: The feature.

    Returns:
      The corresponding SiLA data type.
    """

    if annotation == inspect._empty:
        return Void

    origin = typing.get_origin(annotation) or annotation

    if origin is None:
        return Void
    if origin is typing.Annotated:
        args = typing.get_args(annotation)
        return Constrained.create(data_type=infer(args[0], feature), constraints=list(args[1:]))
    if origin is typing.Union and annotation is Native:
        return Any
    if origin is typing.Any:
        return Any
    if issubclass(origin, type(None)):
        return Void
    if issubclass(origin, CustomDataType):
        return origin.attach(feature)
    if dataclasses.is_dataclass(origin):
        docs = utils.parse_docs(inspect.getdoc(origin))
        fields = dataclasses.fields(origin)

        elements: dict[str, Element] = {}
        fields_by_identifier: dict[str, str] = {}
        for index, field in enumerate(fields):
            field_display_name = utils.humanize(field.name)
            field_identifier = field_display_name.replace(" ", "")
            fields_by_identifier[field_identifier] = field.name
            elements[field.name] = Element(
                identifier=field_identifier,
                display_name=field_display_name,
                description=docs.get("parameter", [])[index].get("default", ""),
                data_type=infer(field.type, feature),
            )

        return Structure.create(elements=elements, name=origin.__name__)

    if issubclass(origin, DataType):
        return origin
    if issubclass(origin, bool):
        return Boolean
    if issubclass(origin, int):
        return Integer
    if issubclass(origin, float):
        return Real
    if issubclass(origin, str):
        return String
    if issubclass(origin, bytes):
        return Binary
    if issubclass(origin, datetime.datetime):
        return Timestamp
    if issubclass(origin, datetime.date):
        return Date
    if issubclass(origin, datetime.time):
        return Time
    if issubclass(origin, list):
        arg = typing.get_args(annotation)
        if not arg:
            msg = f"Unable to identify SiLA type from annotation '{annotation}'"
            raise TypeError(msg)

        return List.create(data_type=infer(arg[0], feature))

    msg = f"Unable to identify SiLA type from annotation '{annotation}'"
    raise TypeError(msg)



# ruff: noqa: D200, D401, D415, E501

from sila import UndefinedExecutionError
from unitelabs.cdk import sila


class TestError(Exception):
    """An error exclusively used for testing purposes"""


class ErrorHandlingTest(sila.Feature):
    """Tests that errors are propagated correctly"""

    def __init__(self):
        super().__init__(
            originator="org.silastandard",
            category="test",
            version="1.0",
        )

    @sila.UnobservableCommand(errors=[TestError])
    async def raise_defined_execution_error(self) -> None:
        """Raises the "Test Error" with the error message 'SiLA2_test_error_message'"""

        msg = "SiLA2_test_error_message"
        raise TestError(msg)

    @sila.ObservableCommand(errors=[TestError])
    async def raise_defined_execution_error_observably(self) -> None:
        """Raises the "Test Error" with the error message 'SiLA2_test_error_message'"""

        msg = "SiLA2_test_error_message"
        raise TestError(msg)

    @sila.UnobservableCommand()
    async def raise_undefined_execution_error(self) -> None:
        """Raises an Undefined Execution Error with the error message 'SiLA2_test_error_message'"""

        msg = "SiLA2_test_error_message"
        raise UndefinedExecutionError(msg)

    @sila.ObservableCommand()
    async def raise_undefined_execution_error_observably(self) -> None:
        """Raises an Undefined Execution Error with the error message 'SiLA2_test_error_message'"""

        msg = "SiLA2_test_error_message"
        raise UndefinedExecutionError(msg)

    @sila.UnobservableProperty(errors=[TestError])
    async def raise_defined_execution_error_on_get(self) -> int:
        """A property that raises a "Test Error" on get with the error message 'SiLA2_test_error_message'"""

        msg = "SiLA2_test_error_message"
        raise TestError(msg)

    @sila.ObservableProperty(errors=[TestError])
    async def raise_defined_execution_error_on_subscribe(self) -> sila.Stream[int]:
        """A property that raises a "Test Error" on subscribe with the error message 'SiLA2_test_error_message'"""

        msg = "SiLA2_test_error_message"
        raise TestError(msg)

    @sila.UnobservableProperty()
    async def raise_undefined_execution_error_on_get(self) -> int:
        """A property that raises an Undefined Execution Error on get with the error message 'SiLA2_test_error_message'"""

        msg = "SiLA2_test_error_message"
        raise UndefinedExecutionError(msg)

    @sila.ObservableProperty()
    async def raise_undefined_execution_error_on_subscribe(self) -> sila.Stream[int]:
        """
        A property that raises an Undefined Execution Error on subscribe with the error message 'SiLA2_test_error_message'
        """

        msg = "SiLA2_test_error_message"
        raise UndefinedExecutionError(msg)

    @sila.ObservableProperty(errors=[TestError])
    async def raise_defined_execution_error_after_value_was_sent(self) -> sila.Stream[int]:
        """
        A property that first sends the integer value 1 and then raises a Defined Execution Error with the error message 'SiLA2_test_error_message'
        """

        yield 1

        msg = "SiLA2_test_error_message"
        raise TestError(msg)

    @sila.ObservableProperty()
    async def raise_undefined_execution_error_after_value_was_sent(self) -> sila.Stream[int]:
        """
        A property that first sends the integer value 1 and then raises a Undefined Execution Error with the error message 'SiLA2_test_error_message'
        """

        yield 1

        msg = "SiLA2_test_error_message"
        raise UndefinedExecutionError(msg)



import collections.abc
import dataclasses
import inspect

import typing_extensions as typing

import sila
from sila.server import Native, SiLAError, UndefinedExecutionError

from ..common import Dataclass
from ..common.errors import define_error
from ..data_types import Any
from ..data_types.convert_data_type import from_sila

if typing.TYPE_CHECKING:
    from ..common import Decorator, Feature

T = typing.TypeVar("T", bound=Any)


@dataclasses.dataclass
class Metadatum(typing.Generic[T], Dataclass):
    """
    Base class for metadata.

    Args:
      identifier: The identifier of the metadata.
      display_name: The display name of the metadata.
      description: The description of the metadata.
    """

    _affects: typing.ClassVar[set[str]] = set()
    _metadatum: typing.ClassVar[typing.Optional[type[sila.server.Metadata]]] = None

    def __init_subclass__(
        cls,
        /,
        *,
        identifier: typing.Optional[str] = None,
        display_name: typing.Optional[str] = None,
        name: typing.Optional[str] = None,
        errors: typing.Optional[collections.abc.Sequence[type[Exception]]] = None,
    ) -> None:
        super().__init_subclass__(identifier=identifier, display_name=display_name, name=name)

        cls._affects = set()
        cls._errors: list[type[Exception]] = list(errors or [])

    @typing.override
    @classmethod
    def attach(cls, feature: "Feature") -> type[sila.server.Metadata]:
        super().attach(feature)

        if cls._identifier in feature.metadata:
            cls._metadatum = typing.cast(type[sila.server.Metadata], feature.metadata[cls._identifier])

            return cls._metadatum

        cls._metadatum = sila.server.Metadata.create(
            identifier=cls._identifier,
            display_name=cls._name,
            description=cls._description,
            errors={Error.identifier: Error for error in cls._errors if (Error := define_error(error))},
            data_type=cls._infer_data_type(feature),
            affects=list(cls._affects),
            function=cls._intercept,
            feature=feature,
        )
        return cls._metadatum

    @classmethod
    async def _intercept(cls, value: Native) -> None:
        """Intercept method execution."""

        try:
            await cls.from_native(value).intercept()
        except SiLAError:
            raise
        except Exception as error:
            if type(error) in cls._errors:
                raise define_error(error)(str(error)) from None

            msg = f"{error.__class__.__name__}: {error}"
            raise UndefinedExecutionError(msg) from error

    @classmethod
    def _infer_metadata(cls, decorator: "Decorator") -> tuple[str, list[type["Metadatum"]]]:
        from .metadata import Metadata

        signature = inspect.signature(decorator._function)

        parameter: str = ""
        metadata: list[type["Metadatum"]] = []

        for param in signature.parameters.values():
            if (
                typing.get_origin(param.annotation) is typing.Annotated
                and (args := typing.get_args(param.annotation))
                and args[0] is Metadata
            ):
                parameter = param.name

                for arg in args[1:]:
                    if not issubclass(arg, Metadatum):
                        msg = (
                            f"Expected instance of `Metadatum` for metadata annotation, "
                            f"received '{arg.__name__}' for parameter '{param.name}'."
                        )
                        raise ValueError(msg)

                    if not decorator._handler:
                        raise RuntimeError

                    arg._affects.add(decorator._handler.fully_qualified_identifier)
                    if arg._metadatum:
                        arg._metadatum.affects = [
                            *arg._metadatum.affects,
                            decorator._handler.fully_qualified_identifier,
                        ]
                    metadata.append(arg)

                break

        return parameter, metadata

    @classmethod
    def from_native(cls, value: Native) -> typing.Self:
        """
        Convert a SiLA metadata value to this counterpart.

        Args:
          value: The value to parse.

        Returns:
          A new instance of this metadatum with the given value.
        """

        if cls._metadatum and issubclass(cls._metadatum.data_type, sila.Structure):
            return cls(**from_sila(value, cls._metadatum.data_type, cls._metadatum.feature))  # type: ignore

        return cls(value)  # type: ignore

    @property
    def _feature(self) -> "Feature":
        """The feature this metadata is registered with."""

        if not self._metadatum or not isinstance(self._metadatum.feature, Feature):
            raise RuntimeError

        return self._metadatum.feature

    async def intercept(self) -> None:
        """Intercept method execution."""



import collections.abc
import dataclasses
import functools
import inspect
import warnings

import typing_extensions as typing

import sila
from sila.server import Server

from .. import utils
from ..metadata import Metadatum
from .decorator import Decorator

if typing.TYPE_CHECKING:
    from ...connector import Connector
    from ..data_types import CustomDataType


@dataclasses.dataclass
class Feature(sila.framework.Feature):
    """
    A feature describes a specific behavior of the server.

    Use the docstring of your feature class to provide a detailed,
    human-readable description of the use of your feature.
    """

    def __init__(
        self,
        *args,
        identifier: typing.Optional[str] = None,
        display_name: typing.Optional[str] = None,
        name: typing.Optional[str] = None,
        description: typing.Optional[str] = None,
        metadata: typing.Optional[collections.abc.Sequence[type[Metadatum]]] = None,
        **kwargs,
    ):
        if display_name is not None:
            msg = "Using `display_name` is deprecated, use `name` instead."
            warnings.warn(msg, stacklevel=2)
            name = display_name

        name = name or utils.to_display_name(self.__class__.__name__)
        identifier = identifier or name.replace(" ", "")
        description = description or next((inspect.getdoc(cls) for cls in inspect.getmro(type(self))), "") or ""

        super().__init__(*args, identifier=identifier, display_name=name, description=description, **kwargs)

        for metadatum in metadata or []:
            metadatum.attach(self).add_to_feature(self)

        self._custom_data_types: dict[str, type["CustomDataType"]] = {}

        self._app: typing.Optional["Connector"] = None

    def attach(self) -> bool:
        """
        Attach all handlers to this feature.

        Returns:
          Whether at least one handler was attached.
        """

        attached = False

        for cls in inspect.getmro(type(self)):
            for name, function in inspect.getmembers(cls, predicate=inspect.isfunction):
                if (handler := getattr(function, "__handler", None)) and isinstance(handler, Decorator):
                    method = getattr(self, name).__func__
                    method = functools.partial(method, self)
                    handler._function = functools.wraps(function)(method)

                    attached = handler.attach(self) or attached

        return attached or bool(self.metadata)

    @property
    def app(self) -> "Connector":
        """The connector app this feature is registered with."""

        if not self._app:
            raise RuntimeError

        return self._app

    @property
    def server(self) -> Server:
        """The server this feature is registered with."""

        if not isinstance(self.context, Server):
            raise RuntimeError

        return self.context



import typing
import weakref

from sila.server import CommandExecution, Element

T = typing.TypeVar("T")


class Intermediate(typing.Generic[T]):
    """A class representing an intermediate response in a command execution."""

    def __init__(self, command_execution: CommandExecution, responses: dict[str, Element]):
        self.command_execution: CommandExecution = weakref.proxy(command_execution)
        self.responses = responses

    def send(self, responses: T, /) -> None:
        """Send an intermediate response."""

        if responses is None:
            return

        result = {}
        responses = [responses] if not isinstance(responses, tuple) else responses
        resps = list(self.responses.values())
        for index, response in enumerate(responses):
            key = resps[index].identifier if index < len(resps) else index
            result[key] = response

        self.command_execution.send_intermediate_responses(result)



# ruff: noqa: D205, D401, D415

from unitelabs.cdk import sila


class UnobservableCommandTest(sila.Feature):
    """Feature for testing unobservable commands"""

    def __init__(self):
        super().__init__(originator="org.silastandard", category="test")

    @sila.UnobservableCommand()
    def command_without_parameters_and_responses(self) -> None:
        """A command that takes no parameters and returns no responses"""

    @sila.UnobservableCommand()
    @sila.Response(name="String Representation")
    def convert_integer_to_string(self, integer: int) -> str:
        """
        A command that takes one integer parameter and returns its string representation.

        .. parameter:: An integer, e.g. 12345
        .. return:: The string representation of the given integer, e.g. '12345'
        """

        return str(integer)

    @sila.UnobservableCommand()
    @sila.Response(name="Joined Parameters")
    def join_integer_and_string(self, integer: int, string: str) -> str:
        """
        A command which takes an integer and a string parameter and returns a string with both joined (e.g.
        "123abc")

        .. parameter:: An integer, e.g. 123
        .. parameter:: A string, e.g. 'abc'
        .. return:: Both parameters joined as string (e.g. '123abc')
        """

        return f"{integer}{string}"

    @sila.UnobservableCommand()
    @sila.Response(name="First Character")
    @sila.Response(name="Remainder")
    def split_string_after_first_character(self, string: str) -> tuple[str, str]:
        """
        A command which splits a given string after its first character. Returns empty parts if the input was
        too short.

        .. parameter:: A string, e.g. 'abcde'
        .. return:: The first character, e.g. 'a', or an empty string if the input was empty
        .. return:: The remainder, e.g. 'bcde', or an empty string if the input was shorter that two characters
        """

        return string[:1], string[1:]



import inspect

import typing_extensions as typing

import sila

from .. import utils
from ..commands import Parameters, Responses
from ..common import Decorator
from ..common.errors import define_error
from ..metadata import Metadatum

if typing.TYPE_CHECKING:
    from ..common import Feature


class UnobservableCommand(Decorator):
    """Any command for which observing the progress of execution is not possible or does not make sense."""

    @typing.override
    def attach(self, feature: "Feature") -> bool:
        if not super().attach(feature):
            return False

        docs = utils.parse_docs(inspect.getdoc(self._function))

        self._name = self._name or utils.humanize(self._function.__name__)
        self._identifier = self._identifier or self._name.replace(" ", "")
        self._description = docs.get("default", "")

        self._parameters = Parameters.from_signature(feature, self._function).elements
        self._responses = Responses.from_signature(feature, self._function).elements

        self._handler = sila.server.UnobservableCommand(
            identifier=self._identifier,
            display_name=self._name,
            description=self._description,
            function=self.execute,
            errors={Error.identifier: Error for error in self._errors if (Error := define_error(error))},
            parameters=self._parameters,
            responses=self._responses,
            feature=feature,
        )
        self._metadata = Metadatum._infer_metadata(self)

        return True



import dataclasses
import inspect
import typing
import warnings

import sila

from .. import utils


def define_error(exception: typing.Union[Exception, type[Exception]]) -> type[sila.DefinedExecutionError]:
    """
    Convert an exception into a defined execution error.

    Args:
      exception: The exception class or instance to convert.

    Returns:
      A DefinedExecutionError object with the parsed information from the exception.
    """

    if not inspect.isclass(exception):
        exception = exception.__class__

    if issubclass(exception, sila.DefinedExecutionError):
        return exception

    display_name = utils.to_display_name(exception.__name__)
    identifier = display_name.replace(" ", "")
    description = inspect.getdoc(exception) or ""

    return sila.DefinedExecutionError.create(identifier=identifier, display_name=display_name, description=description)


@dataclasses.dataclass
class DefinedExecutionError(Exception):
    """A defined execution error."""

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()

        msg = "Using `sila.DefinedExecutionError` is deprecated, please directly inherit from `Exception`."
        warnings.warn(msg, stacklevel=2)

    def __init__(self, *args, identifier: str = "", display_name: str = "", description: str = "", **kwargs):
        pass



import collections.abc
import inspect
import warnings

import typing_extensions as typing

import sila
from sila.server import Native, SiLAError, UndefinedExecutionError
from unitelabs.cdk.subscriptions import Subscription

from .. import utils
from ..common import Decorator
from ..common.errors import define_error
from ..data_types import infer, to_sila
from ..metadata import Metadatum

if typing.TYPE_CHECKING:
    from ..common import Feature

T = typing.TypeVar("T")
Stream = collections.abc.AsyncIterator[T]


class ObservableProperty(Decorator):
    """A property describes certain aspects of a SiLA server that do not require an action on the SiLA server."""

    @typing.override
    def attach(self, feature: "Feature") -> bool:
        if not super().attach(feature):
            return False

        self._name = self._name or utils.humanize(self._function.__name__.removeprefix("subscribe_"))
        self._identifier = self._identifier or self._name.replace(" ", "")
        self._description = inspect.getdoc(self._function) or ""

        type_hint = return_annotation = inspect.signature(self._function).return_annotation
        if type_hint is inspect._empty:
            type_hint = Stream[Native]

        type_hint = typing.get_args(type_hint)

        if not type_hint:
            output = return_annotation.__name__ if hasattr(return_annotation, "__name__") else return_annotation
            warnings.warn(
                f"Invalid return type for observable property '{self._identifier}'"
                f": expected `sila.Stream[...]`, received `{output}`.'",
                stacklevel=2,
            )
            type_hint = (return_annotation,)

        data_type = infer(type_hint[0], feature)

        self._responses = {
            self._identifier: sila.Element(identifier=self._identifier, display_name=self._name, data_type=data_type)
        }

        self._handler = sila.server.ObservableProperty(
            identifier=self._identifier,
            display_name=self._name,
            description=self._description,
            function=self.execute,
            errors={Error.identifier: Error for error in self._errors if (Error := define_error(error))},
            data_type=data_type,
            feature=feature,
        )
        self._metadata = Metadatum._infer_metadata(self)

        return True

    @typing.override
    async def execute(
        self, metadata: dict[sila.MetadataIdentifier, Native], **parameters
    ) -> collections.abc.AsyncIterator[Native]:
        if not self._feature:
            raise RuntimeError

        try:
            function = self._with_metadata(self._function, metadata)
            function = self._with_parameters(function, parameters)

            responses = self._execute(function)

            async for item in responses:
                yield to_sila(item, sila.Structure.create(self._responses), self._feature)
        except SiLAError:
            raise
        except Exception as error:
            if type(error) in self._errors:
                raise define_error(error)(str(error)) from None

            msg = f"{error.__class__.__name__}: {error}"
            raise UndefinedExecutionError(msg) from error

    @typing.override
    async def _execute(self, function: collections.abc.Callable) -> collections.abc.AsyncIterator[Native]:
        responses = function()

        if inspect.iscoroutine(responses):
            responses = await responses

        if isinstance(responses, Subscription):
            async for response in responses:
                yield {self._identifier: response}

            responses.terminate()

        elif inspect.isasyncgen(responses):
            async for response in responses:
                yield {self._identifier: response}

        elif inspect.isgenerator(responses):
            for response in responses:
                yield {self._identifier: response}



import typing

from sila import FeatureIdentifier
from unitelabs.cdk.features.core import authentication_service


class AuthenticationService(authentication_service.AuthenticationService):
    @typing.override
    async def validate(self, username: str, password: str, scope: list[FeatureIdentifier]) -> bool:
        return (username, password) == ("test", "test")



# ruff: noqa: D205, D401, D415, E501

import dataclasses

from unitelabs.cdk import sila


@dataclasses.dataclass
class TestStructure(sila.CustomDataType):
    """
    An example Structure data type containing all SiLA basic types.

    .. parameter:: A value of SiLA data type String.
    .. parameter:: A value of SiLA data type Integer.
    .. parameter:: A value of SiLA data type Real.
    .. parameter:: A value of SiLA data type Boolean.
    .. parameter:: A value of SiLA data type Binary.
    .. parameter:: A value of SiLA data type Date.
    .. parameter:: A value of SiLA data type Time.
    .. parameter:: A value of SiLA data type Timestamp
    .. parameter:: A value of SiLA data type Any.
    """

    string_type_value: str
    integer_type_value: int
    real_type_value: float
    boolean_type_value: bool
    binary_type_value: bytes
    date_type_value: sila.datetime.date
    time_type_value: sila.datetime.time
    timestamp_type_value: sila.datetime.datetime
    any_type_value: sila.Any


class ListDataTypeTest(sila.Feature):
    """Provides commands and properties to set or respectively get SiLA List Data Type values via command parameters or property responses respectively."""

    def __init__(self):
        super().__init__(originator="org.silastandard", category="test")

    @sila.UnobservableProperty()
    def empty_string_list(self) -> list[str]:
        """Returns an empty list of String type."""

        return []

    @sila.UnobservableCommand()
    @sila.Response(name="Received Values")
    def echo_string_list(self, string_list: list[str]) -> list[str]:
        """
        Receives a list of string values and returns a list containing the string values that have been received.

        .. parameter:: The list of string values to be returned.
        .. return:: A list containing the string values that have been received.
        """

        return string_list

    @sila.UnobservableProperty()
    def string_list(self) -> list[str]:
        """Returns a list with the following string values: 'SiLA 2', 'is', 'great'."""

        return ["SiLA 2", "is", "great"]

    @sila.UnobservableCommand()
    @sila.Response(name="Received Values", description="A list containing the integer values that have been received.")
    def echo_integer_list(self, integer_list: list[int]) -> list[int]:
        """
        Receives a list of integer values and returns a list containing the integer values that have been received.

        .. parameter:: The list of integer values to be returned.
        .. return:: A list containing the integer values that have been received.
        """

        return integer_list

    @sila.UnobservableProperty()
    def integer_list(self) -> list[int]:
        """Returns a list with the following Integer values: 1, 2, 3."""

        return [1, 2, 3]

    @sila.UnobservableCommand()
    @sila.Response(name="ReceivedValues")
    def echo_structure_list(self, structure_list: list[TestStructure]) -> list[TestStructure]:
        """
        Receives a list of structure values and returns a list containing the structures that have been received.

        .. parameter:: The list of structure values to be returned.
        .. return:: A message containing the content of all structures that have been received.
        """

        return structure_list

    @sila.UnobservableProperty()
    def structure_list(self) -> list[TestStructure]:
        """
        Returns a list with 3 structure values, whereas the values of the first element are:
        string value = 'SiLA2_Test_String_Value_1'
        integer value = 5124
        real value = 3.1415926
        boolean value = true
        binary value (embedded string) = 'Binary_String_Value_1'
        date value = 05.08.2022 respective 08/05/2022
        time value = 12:34:56.789
        time stamp value = 2022-08-05 12:34:56.789
        any type value (string) = 'Any_Type_String_Value_1'

        For the second and third element:
        the last character of the strings changes to '2' respective '3'
        the numeric values are incremented by 1
        the boolean values becomes false for element 2 and true for element 3
        for the date value day, month and year are incremented by 1
        for the time value milliseconds, seconds, minutes and hours are incremented by 1
        for the time stamp value day, month, year, milliseconds, seconds, minutes and hours are incremented by 1.
        """

        return [
            TestStructure(
                string_type_value="SiLA2_Test_String_Value_1",
                integer_type_value=5124,
                real_type_value=3.1415926,
                boolean_type_value=True,
                binary_type_value=b"Binary_String_Value_1",
                date_type_value=sila.datetime.date(
                    year=2022, month=8, day=5, tzinfo=sila.datetime.timezone(offset=sila.datetime.timedelta(hours=+2))
                ),
                time_type_value=sila.datetime.time(
                    hour=12,
                    minute=34,
                    second=56,
                    microsecond=789000,
                    tzinfo=sila.datetime.timezone(offset=sila.datetime.timedelta(hours=+2)),
                ),
                timestamp_type_value=sila.datetime.datetime(
                    year=2022,
                    month=8,
                    day=5,
                    hour=12,
                    minute=34,
                    second=56,
                    microsecond=789000,
                    tzinfo=sila.datetime.timezone(offset=sila.datetime.timedelta(hours=+2)),
                ),
                any_type_value="Any_Type_String_Value_1",
            ),
            TestStructure(
                string_type_value="SiLA2_Test_String_Value_2",
                integer_type_value=5125,
                real_type_value=4.1415926,
                boolean_type_value=False,
                binary_type_value=b"Binary_String_Value_2",
                date_type_value=sila.datetime.date(
                    year=2023, month=9, day=6, tzinfo=sila.datetime.timezone(offset=sila.datetime.timedelta(hours=+2))
                ),
                time_type_value=sila.datetime.time(
                    hour=13,
                    minute=35,
                    second=57,
                    microsecond=790000,
                    tzinfo=sila.datetime.timezone(offset=sila.datetime.timedelta(hours=+2)),
                ),
                timestamp_type_value=sila.datetime.datetime(
                    year=2023,
                    month=9,
                    day=6,
                    hour=13,
                    minute=35,
                    second=57,
                    microsecond=790000,
                    tzinfo=sila.datetime.timezone(offset=sila.datetime.timedelta(hours=+2)),
                ),
                any_type_value="Any_Type_String_Value_2",
            ),
            TestStructure(
                string_type_value="SiLA2_Test_String_Value_3",
                integer_type_value=5126,
                real_type_value=5.1415926,
                boolean_type_value=True,
                binary_type_value=b"Binary_String_Value_3",
                date_type_value=sila.datetime.date(
                    year=2024,
                    month=10,
                    day=7,
                    tzinfo=sila.datetime.timezone(offset=sila.datetime.timedelta(hours=+2)),
                ),
                time_type_value=sila.datetime.time(
                    hour=14,
                    minute=36,
                    second=58,
                    microsecond=791000,
                    tzinfo=sila.datetime.timezone(offset=sila.datetime.timedelta(hours=+2)),
                ),
                timestamp_type_value=sila.datetime.datetime(
                    year=2024,
                    month=10,
                    day=7,
                    hour=14,
                    minute=36,
                    second=58,
                    microsecond=791000,
                    tzinfo=sila.datetime.timezone(offset=sila.datetime.timedelta(hours=+2)),
                ),
                any_type_value="Any_Type_String_Value_3",
            ),
        ]



import datetime

from unitelabs.cdk import sila


class GreetingProvider(sila.Feature):
    """
    Example implementation of a minimum Feature.

    Provides a Greeting to the Client and a StartYear property, informing about the year the Server has been started.
    """

    def __init__(self):
        super().__init__(
            originator="org.silastandard",
            category="examples",
            version="1.0",
            maturity_level="Verified",
        )

        self._start_year = datetime.datetime.now().year

    @sila.UnobservableCommand()
    @sila.Response(name="Greeting")
    async def say_hello(self, name: str) -> str:
        """
        Say "Hello SiLA 2 + [Name]" to the client.

        .. parameter:: The name, SayHello shall use to greet.
        .. return: The greeting string, returned to the SiLA Client.
        """

        return f"Hello SiLA 2 {name}"

    @sila.UnobservableProperty()
    async def start_year(self) -> int:
        """Get the year the SiLA Server has been started in."""

        return self._start_year



import asyncio
import typing

import pytest

ONE_OP_TIME = 0.01
DEFAULT_VALUE = "default"


async def bg_cancel(cancel_event: asyncio.Event):
    """Set the cancel event after a delay of ~5 operations."""
    await asyncio.sleep(ONE_OP_TIME * 5)
    cancel_event.set()


@pytest.fixture
def create_task() -> typing.Generator[asyncio.Task, None, None]:
    tasks = set()

    def _create_task(
        method: typing.Coroutine[typing.Awaitable, None, None],
    ) -> typing.Generator[asyncio.Task, None, None]:
        name = f"subscription-{method.__name__}"
        task = asyncio.create_task(method, name=name)
        tasks.add(task)
        task.add_done_callback(tasks.discard)

        yield task

        task.cancel()
        tasks.discard(task)

    yield _create_task

    for task in tasks:
        task.cancel()
        tasks.discard(task)



import typing
import unittest.mock

import pytest

import sila
from sila.framework.errors.defined_execution_error import DefinedExecutionError
from unitelabs.cdk.sila.common.feature import Feature
from unitelabs.cdk.sila.property.observable_property import ObservableProperty, Stream


class DefinedError(Exception):
    pass


@pytest.fixture
def feature():
    return Feature(identifier="Feature", name="Feature")


@pytest.fixture
def handler():
    return ObservableProperty(identifier="ObservableProperty", name="Observable Property", errors=[DefinedError])


class TestAttach:
    async def test_should_attach_handler(self, feature: Feature, handler: ObservableProperty):
        # Attach
        attached = handler.attach(feature)

        # Assert that the method returns the correct value
        assert attached is True
        assert feature.properties[handler._identifier] == handler._handler

    async def test_should_ignore_disabled_handler(self, feature: Feature, handler: ObservableProperty):
        # Attach
        handler._enabled = False
        attached = handler.attach(feature)

        # Assert that the method returns the correct value
        assert attached is False
        assert handler._identifier not in feature.properties

    async def test_should_call_enabled_callback(self, feature: Feature, handler: ObservableProperty):
        # Attach
        handler._enabled = unittest.mock.Mock(return_value=True)
        attached = handler.attach(feature)

        # Assert that the method returns the correct value
        assert attached is True
        assert feature.properties[handler._identifier] == handler._handler
        handler._enabled.assert_called_once_with(feature)

    async def test_should_attach_return_annotation(self, feature: Feature, handler: ObservableProperty):
        # Create handler
        def function() -> Stream[int]:
            return unittest.mock.sentinel.response

        handler(function)

        # Attach
        handler.attach(feature)

        # Assert that the method returns the correct value
        assert isinstance(handler._handler, sila.ObservableProperty)
        assert handler._handler.data_type == sila.Integer

    async def test_should_attach_missing_return_annotation(self, feature: Feature, handler: ObservableProperty):
        # Create handler
        def function():
            return unittest.mock.sentinel.response

        handler(function)

        # Attach
        handler.attach(feature)

        # Assert that the method returns the correct value
        assert isinstance(handler._handler, sila.ObservableProperty)
        assert handler._handler.data_type == sila.Any

    async def test_should_attach_annotated_return(self, feature: Feature, handler: ObservableProperty):
        # Create handler
        def function() -> Stream[typing.Annotated[int, sila.Unit("s", [sila.UnitComponent("Second")])]]:
            return unittest.mock.sentinel.response

        handler(function)

        # Attach
        handler.attach(feature)

        # Assert that the method returns the correct value
        assert isinstance(handler._handler, sila.ObservableProperty)
        assert issubclass(handler._handler.data_type, sila.Constrained)
        assert handler._handler.data_type.data_type == sila.Integer
        assert handler._handler.data_type.constraints == [sila.Unit("s", [sila.UnitComponent("Second")])]


class TestExecute:
    # Execute synchronous function with default parameters.
    async def test_execute_synchronous_default_parameters(self, feature: Feature, handler: ObservableProperty):
        # Initialize the observable property
        callback = unittest.mock.Mock()

        async def function() -> Stream[int]:
            callback()
            yield 1
            yield 2
            yield 3

        handler(function)
        handler.attach(feature)

        # Execute function
        result = handler.execute(metadata={})
        result_0 = await result.__anext__()
        result_1 = await result.__anext__()
        result_2 = await result.__anext__()

        # Assert that the function was called with the correct arguments
        callback.assert_called_once_with()

        # Assert that the method returns the correct value
        assert result_0 == {"ObservableProperty": 1}
        assert result_1 == {"ObservableProperty": 2}
        assert result_2 == {"ObservableProperty": 3}

    # Verify that the method raises an error when the synchronous function raises.
    async def test_raises_when_synchronous_raises(self, feature: Feature, handler: ObservableProperty):
        # Initialize the observable property
        def function() -> Stream[int]:
            msg = "Hello, World!"
            raise Exception(msg)

        handler(function)
        handler.attach(feature)

        # Execute function
        result = handler.execute(metadata={})
        with pytest.raises(Exception, match=r"Exception: Hello, World!"):
            await result.__anext__()

    # Verify that the method raises a defined execution error when the synchronous decorator knows the error type.
    async def test_raises_when_synchronous_raises_known_error(self, feature: Feature, handler: ObservableProperty):
        # Initialize the observable property
        def function() -> Stream[int]:
            msg = "Hello, World!"
            raise DefinedError(msg)

        handler(function)
        handler.attach(feature)

        # Execute function
        result = handler.execute(metadata={})
        with pytest.raises(DefinedExecutionError) as exc_info:
            await result.__anext__()

        assert exc_info.value.identifier == "DefinedError"
        assert exc_info.value.display_name == "Defined Error"
        assert exc_info.value.description == "Common base class for all non-exit exceptions."
        assert exc_info.value.message == "Hello, World!"

    # Execute asynchronous function with default parameters.
    async def test_execute_asynchronous_default_parameters(self, feature: Feature, handler: ObservableProperty):
        # Initialize the observable property
        callback = unittest.mock.AsyncMock()

        async def function() -> Stream[int]:
            await callback()
            yield 1
            yield 2
            yield 3

        handler(function)
        handler.attach(feature)

        # Execute function
        result = handler.execute(metadata={})
        result_0 = await result.__anext__()
        result_1 = await result.__anext__()
        result_2 = await result.__anext__()

        # Assert that the function was called with the correct arguments
        callback.assert_awaited_once_with()

        # Assert that the method returns the correct value
        assert result_0 == {"ObservableProperty": 1}
        assert result_1 == {"ObservableProperty": 2}
        assert result_2 == {"ObservableProperty": 3}

    # Verify that the method raises an error when the asynchronous function raises.
    async def test_raises_when_asynchronous_raises(self, feature: Feature, handler: ObservableProperty):
        # Initialize the observable property
        async def function() -> Stream[int]:
            msg = "Hello, World!"
            raise Exception(msg)

        handler(function)
        handler.attach(feature)

        # Execute function
        result = handler.execute(metadata={})
        with pytest.raises(Exception, match=r"Exception: Hello, World!"):
            await result.__anext__()

    # Verify that the method raises a defined execution error when the asynchronous decorator knows the error type.
    async def test_raises_when_asynchronous_raises_known_error(self, feature: Feature, handler: ObservableProperty):
        # Initialize the observable property
        async def function() -> Stream[int]:
            msg = "Hello, World!"
            raise DefinedError(msg)

        handler(function)
        handler.attach(feature)

        # Execute function
        result = handler.execute(metadata={})
        with pytest.raises(DefinedExecutionError) as exc_info:
            await result.__anext__()

        assert exc_info.value.identifier == "DefinedError"
        assert exc_info.value.display_name == "Defined Error"
        assert exc_info.value.description == "Common base class for all non-exit exceptions."
        assert exc_info.value.message == "Hello, World!"
